# audioGuide copyright Benjamin Hackbarth, 2011-2013 #
__author__ = "Benjamin Hackbarth, Norbert Schnell, Philippe Esling, Diemo Schwarz, Gilbert Nouno"
__author_email__ = "b.hackbarth@liverpool.ac.uk"
__version__ = "1.03.4"

# import standard libraries/functions
import sys, os, random, operator, tempfile
from time import clock, localtime
from copy import deepcopy
# json
try:
	import json as json
except ImportError:
	try:
		import simplejson as json
	except: pass
# numpy
try:
	import numpy as np
except ImportError:
	userinterface.lady_print( 'AudioGuide needs the `numpy\' python module.  Download and install -- http://docs.scipy.org/doc/numpy/user/install.html' )
	sys.exit(1)



import bin_config
config = bin_config.get(__version__, os.path.dirname(__file__))
# lib dur for PURE PYTHON MODULES
sys.path.append(os.path.join(os.path.dirname(__file__), 'pylib'))
# lib dur for PRECOMPILED MODULES

sys.path.append(config['py_compiled_lib_dir'])
import userinterface
import parseops
import plot
import metrics
import normalize as norm
import simcalc
import utilities
import sdiflinkage
# include audioguide files
execfile(os.path.join(os.path.dirname(__file__), 'classes.py'))


class main:
	def __init__(self, scriptPath, defaultOptionsPath, loadOptionsFromFile=None, pythonicUserOptions={}, printName=None, optionsPath=None):
		global ag, tgt, outputFiles, ops, settingsFile, src, defaults, p
		agPath = os.path.dirname(__file__)
		ops = parseops.parseOptions(agPath, scriptPath, loadOptionsFromFile, pythonicUserOptions, defaultOptionsPath)
		#if ops.LOG_FILEPATH == None: log = None
		#else: log = ops.LOG_FILEPATH
		p = userinterface.Printer(__version__, ops.VERBOSITY, ops.USE_PROGRESS_BAR, ops.PRINT_LENGTH, ops.LOG_FILEPATH)
		p.middle(printName, 2)
		for module in [simcalc, userinterface, metrics, norm]: 
			setattr(module, 'ops', globals()['ops']) # add ops to module's namespace
			setattr(module, 'p', globals()['p']) # add p.post to module's namespace






	def init(self, printConfig=True):
		global ag, tgt, outputFiles, ops, settingsFile, src, defaults, agData, SdifInterface, descriptorData

		SdifInterface = sdiflinkage.SdifInterface(ops, config['bin_path']['ircamd'], config['bin_path']['pm2'], config['bin_path']['supervp'], winLengthSec=ops.DESCRIPTOR_WIN_SIZE_SEC, hopLengthSec=ops.DESCRIPTOR_HOP_SIZE_SEC, resampleRate=ops.IRCAMDESCRIPTOR_RESAMPLE_RATE, windowType=ops.IRCAMDESCRIPTOR_WINDOW_TYPE, numbMfccs=ops.IRCAMDESCRIPTOR_NUMB_MFCCS, F0MaxAnalysisFreq=ops.IRCAMDESCRIPTOR_F0_MAX_ANALYSIS_FREQ, F0MinFrequency=ops.IRCAMDESCRIPTOR_F0_MIN_FREQUENCY, F0MaxFrequency=ops.IRCAMDESCRIPTOR_F0_MAX_FREQUENCY, F0AmpThreshold=ops.IRCAMDESCRIPTOR_F0_AMP_THRESHOLD, F0Quality=ops.IRCAMDESCRIPTOR_F0_QUALITY, numbPeaks=ops.SUPERVP_NUMB_PEAKS, numbClust=ops.CLUSTERANAL_NUMB_CLUSTS, clustDescriptDict=ops.CLUSTERANAL_DESCRIPTOR_DIM, forceAnal=ops.DESCRIPTOR_FORCE_ANALYSIS, p=p)
		
		descriptorData = DescriptorManager()
		
		descriptorData.allDescriptorsList(SdifInterface) # pass reference
		
		metrics.set_all_segmented_methods(descriptorData.allSegmentMethods)

		ag = analData()
		
		for module in [simcalc, userinterface]: setattr(module, 'ag', globals()['ag']) # add ag to each module's namespace
		src = source(ops.CORPUS)
		agData = otherSavedData(ops.JSON_DATA_FILEPATH, "audioguide-data")
		p.post( "" )
		ops.needToInitalize = False




	def changeOption(self, name, value):
		global ops
		setattr(ops, name, value)
	
	
	



	def loadTarget(self):
		global ag, tgt, outputFiles, ops, settingsFile, src, descriptorData
		p.logsection('TARGET')
		if ops.needToInitalize:
			self.init()
		ag.timer("target", "set")
		tgt = targetData(ops.TARGET.name, ops.TARGET.start, ops.TARGET.end, ops.TARGET.thresh, ops.TARGET.rise, ops.TARGET.scaleDb, ops.TARGET.minSegLen, ops.TARGET.maxSegLen, config['bin_path']['supervp']) # load the target soundfile; do metrics
		printStr = '%s from %.2f->%.2f'%(tgt.filehead, tgt.startSec, tgt.endSec)
		p.barOpen('Evaluating Target '+printStr, 4)
		p.post(printStr, inc=True)
		tgt.initAnal(descriptorData)
		tgt.data['orig'] = deepcopy(tgt.data['raw']) # preserve a copy of original data
		tgt.onsetInit()
		tgt.offsetInit()
		tgt.initMixture()		
		tgt.segs = tgt.getSegs() # dont need first times
		tgt.initSegMethod(descriptorData.target_seg)
		p.barClose("Done.")
		if ops.TGT_OM_SCORE_FILEPATH != None:
			tgt.writeOMScore(ops.TGT_OM_SCORE_FILEPATH, tgt.segs)
		if ops.TARGET_SEGMENT_LABELS_FILEPATH != None:
			tgt.writeSegLabelsToDisk(ops.TARGET_SEGMENT_LABELS_FILEPATH, tgt.segs, ops.TEMPO_CHANGE)
		if ops.TARGET_PLOT_METRICS:
			plot.descriptorPlots(ops.PLOT_PATH, tgt.data['raw'], tgt.segFrames, ops.DESCRIPTOR_HOP_SIZE_SEC)
		ops.needToLoadTgt = False		
		ops.needToRewindTgt = False
		ag.timer("target", "get")



	def loadDatabase(self):
		global ag, tgt, outputFiles, ops, settingsFile, src, p, descriptorData
		if ops.needToInitalize: self.init()
		startTime = clock()
		segmentFileDetect = ''
		ag.timer("corpus", "set")
		p.barOpen("Evaluating Corpus...", len(ag.preLimitCpsHandles))
		# init analysis for each corpus segment that passed user CORPUS restrictions (except limits)
		if len(ag.preLimitCpsHandles) == 0:
			utilities.error('CORPUS', "No segments made it into the CORPUS")
		for sf in ag.preLimitCpsHandles:
			sf.initAnal(descriptorData)
		########################
		# LIMITS - test
		# find min/max of descriptor variance if using a percentage
		for cpsLimitObj in ag.limitObjList:
			cpsLimitObj.checkminMax(ag.preLimitCpsHandles)
		# test
		for sf in ag.preLimitCpsHandles:
			passed = []
			for limitObj in sf.limit:
				test = limitObj.test(sf)
				passed.append(test)
			if False in passed: continue # then ignore this segment!
			# add the segment since if passed any user-supplied limitations...
			ag.allHandles.append(sf)
			ag.sourceHandles.append(sf)
			src.pool[sf.voiceID][sf] = 0
			ag.srcMaxPowers.append(max(sf.data['raw']['power'])) # in case power-seg is not max!
			ag.corpusLengthInSeconds += sf.endSec-sf.startSec
		p.barClose("Read %i/%i segments (%.0f%%, %.2f min.)"%(len(ag.sourceHandles), len(ag.preLimitCpsHandles), len(ag.sourceHandles)/float(len(ag.preLimitCpsHandles))*100., ag.corpusLengthInSeconds/60.))
		
		if len(ag.limitObjList) > 0:
			print('LIMITS')
			for cpsLimitObj in ag.limitObjList:
				print( cpsLimitObj.printRejects() )
			print('\n')

		# test to make sure some samples made it though... 
		if len(ag.sourceHandles) == 0:
			utilities.error('CORPUS', "No database segments made it into the selection pool.  Check limits..")
		if len(ops.ORDER_CORPUS_BY_DESCRIPTOR) > 0: self.renderCorpusOrderSfAndLabels()
		ops.needToLoadSrc = False	
		ag.timer("corpus", "get")

		




	def renderCorpusOrderSfAndLabels(self, tempo=144, gapBetweenSounds=0.2):
		for descriptor in ops.ORDER_CORPUS_BY_DESCRIPTOR:
			tmp = []
			for h in ag.sourceHandles: tmp.append((h.data['raw'][descriptor][0], h))
			tmp.sort() # get ordering
			scoTime = 0
			notes = []
			labels = []
			allTimes = []
			for val, h in tmp:
				allTimes.append((h.startSec, h.endSec, h.filename))
				dur = (h.endSec-h.startSec)
				notes.append(['i1', str(scoTime), str(dur), "100", '"'+os.path.abspath(h.filename)+'"', str(h.startSec), "1", str(utilities.ampToDb(h.data['raw']['power-seg'][0])), str(h.data['raw']['peakTime-seg'][0]), str(h.segdata.getraw('effDur-seg', 0, None)), '0.01', '0.01', str(int(h.voiceID)), '0'])
				labels.append((str(scoTime), str(scoTime+dur), str(h.data['raw'][descriptor][0]), h.name))
				if tempo == 'cpsDur':
					scoTime += dur+gapBetweenSounds
				elif type(tempo) in [float, int]:
					scoTime += 60./float(tempo)
			
			tmpsco = tempfile.NamedTemporaryFile('w')
			tmplabels = os.path.join(os.path.split(ops.CSOUND_RENDER_FILEPATH)[0], 'corpus-ordered-'+descriptor+'.txt')
			labelsh = open(tmplabels, 'w')
			for note in notes:
				scoh.write(' '.join(note)+'\n')
			for label in labels:
				labelsh.write('\t'.join(label)+'\n')
			labelsh.close()
			outputPath = os.path.join(os.path.split(ops.CSOUND_RENDER_FILEPATH)[0], 'corpus-ordered-'+descriptor+'.aiff')
			renderWithCsound([csoundRenderObj(tmpsco.name, outputPath, len(notes))], "Rendering corpus ordered by "+descriptor)







	def normalizeData(self):
		global ag, tgt, outputFiles, ops, settingsFile, src, agData, SdifInterface
		p.logsection('NORMALIZATION')
		p.post("Normalizing Data By", ops.NORMALIZE_METHOD,"..", 4)
		ag.timer("norm", "set")

#		for d in descriptorData.master_list:
#			if d.norm == None or d.weight <= 0: continue
#			d.get_norm_coeffs(tgt, ag.sourceHandles, 'raw')
#			# make norm data dict entries
#			if d.seg:
#				for h in ag.allHandles:
#					h.data['norm'][d.name] = [None] * h.lengthInFrames
#			else:
#				for h in ag.allHandles:
#					h.data['norm'][d.name] = np.empty(h.lengthInFrames)
#
#			norm.normArray(tgt, d, d.tgt_coeff, 'raw', 'norm', 0, None)
#			for h in ag.sourceHandles:
#				norm.normArray(h, d, d.cps_coeff, 'raw', 'norm', 0, None)
		
		#sys.exit()

		# normalize raw data..		
		norm_compute = []
		for d in descriptorData.master_list:
			if d.norm != None: norm_compute.append(d)		
		
		if ops.NORMALIZE_METHOD in ['minmax', 'stddev']:
			ag.targetNormCoeffs, ag.sourceNormCoeffs = norm.normalizeCoefs(norm_compute, tgt, ag, p, ops.NORMALIZE_METHOD)

			norm.normalizeDescriptors([tgt], norm_compute, 'raw', 'norm', ops.NORMALIZE_METHOD, ag.targetNormCoeffs, None, None) 
			norm.normalizeDescriptors(ag.sourceHandles, norm_compute, 'raw', 'norm', ops.NORMALIZE_METHOD, ag.sourceNormCoeffs, None, None) 
		# normalize with clustering algorithm
		elif ops.NORMALIZE_METHOD == 'timbreRemap':		
			# now reevaluate the target's normalized space according to the corpus' distribution
			evalTgtSegments, notRemapDescripts = norm.targetTimbreReassignment(descriptorData, tgt, ag.sourceHandles, 'raw', agData)
			tgt.initSegMethod(evalTgtSegments)
			# normalize all metrics with the same coeffs (now ignore norm #2)
			ag.targetNormCoeffs, ag.sourceNormCoeffs = norm.normalizeCoefs(norm_compute, tgt, ag, p, 'stddev')
			allNormCoeffs = {}
			for d in norm_compute:
				allNormCoeffs[d.name] = norm.getNormCoefs(ag.allHandles, d, None, None, 'stddev')

			norm.normalizeDescriptors([tgt], norm_compute, 'raw', 'norm', 'stddev', allNormCoeffs, None, None) 
			norm.normalizeDescriptors(ag.sourceHandles, norm_computee, 'raw', 'norm', 'stddev', allNormCoeffs, None, None) 

#		# normalize with clustering algorithm
#		elif ops.NORMALIZE_METHOD == 'cluster':			
#			norm.normalizeByCluster(analMetrics, tgt, ag, agData)
#		elapsedTime = ag.timer("norm", "get")
#		agData.close()
#		ops.needToNormalize = False		








	def concatenate(self):
		global ag, tgt, outputFiles, ops, settingsFile, src, descriptorData
		# each of these checks insure that the needed analysis routines have been run before concatenation
		if ops.needToInitalize:
			self.init()
		if ops.needToRemakeAnalLists: descriptorData = DescriptorManager()			
		if ops.needToLoadTgt: # need to initalize the target soundfile?
			self.loadTarget()
		if ops.needToLoadSrc: # need to init and load the corpus?
			self.loadDatabase()
		if ops.needToNormalize:
			self.normalizeData()
		if ops.needToRewindTgt: # need to rewind target -- i.e. restore the data for an additional concate call?
			tgt.rewind()
		random.seed(ops.RANDOM_SEED)
		outputFiles = outputData(ops.CSOUND_SCORE_FILEPATH, ops.QLIST_FILEPATH, tgt, ops.CSOUND_RENDER_FILEPATH, ops.OM_SCORE_FILEPATH, ops.MIDI_FILEPATH)
		src.initConcatenate(tgt.lengthInFrames, ag.sourceHandles)
		src.initTimeVaryingRestrict()
		ag.timer("concat", "set")
		tgt.nextIncrementLength = 1
		######################
		p.logsection('SEARCH')
		searchObj = simcalc.distanceCalculations(descriptorData)
		self.NEWNEWconcatenatePower(searchObj)
		if ops.SEARCH_CORPUS_PARAM_INTERPOLATE:
			searchObj.paramInterpolateScore(os.path.split(ops.CSOUND_SCORE_FILEPATH)[0])
		searchObj.close()
		#######################
		elapsedTime = np.clip(ag.timer("concat", "get"), 0.0001, sys.maxint)
		p.barClose("Picked "+utilities.bold(str(outputFiles.totalOutputNotes))+" notes in "+str(utilities.trunc(elapsedTime, 2))+" sec. ("+str(utilities.trunc((ag.f2s(tgt.lengthInFrames))/elapsedTime, 2))+"x realtime)")
		outputFiles.done(ops.CSOUND_SCORE_FILEPATH, ops.QLIST_FILEPATH, ops.OM_SCORE_FILEPATH, ops.MIDI_FILEPATH)
		if True:
			sumTerminations = 0.0
			for key in ag.logicForSegmentTermination: sumTerminations += ag.logicForSegmentTermination[key]
			if sumTerminations != 0:
				for key in ag.logicForSegmentTermination: ag.logicForSegmentTermination[key] = (ag.logicForSegmentTermination[key]/sumTerminations)*100
				utilities.printDict("segment termination logic", ag.logicForSegmentTermination, 2, p)

		if ops.PRINT_SELECTION_HISTO: src.printSampleHistogram()
		if ops.PRINT_SIM_SELECTION_HISTO: src.printSimultanousHistogram()
		# other output files
		if ops.SUPERIMPOSITION_LABEL_FILEPATH != None:
			tgt.writeSegLabelsToDisk(ops.SUPERIMPOSITION_LABEL_FILEPATH, tgt.selectedSegs, ops.TEMPO_CHANGE)
		if ops.OM_SCORE_FILEPATH != None:
			outputFiles.writeOMScoreNote()
		if ops.BACH_SCORE_FILEPATH != None:
			outputFiles.writeBachScoreNotes(ops.BACH_SCORE_FILEPATH, ag.sourceHandles, ops.CSOUND_RENDER_FILEPATH)
		if ops.LISP_FILEPATH != None:
			outputFiles.writeLispEvents(ops.LISP_FILEPATH)
		ops.needToRewindTgt = True
		#for data in searchObj.searchResults:
		#	print data








	def NEWNEWconcatenatePower(self, searchObj):
		global ag, tgt, outputFiles, ops, settingsFile, src
		error_accum = 0.
		p.barOpen('Concatenating Based on %s'%ops.SUPERIMPOSE.searchOrder, len(tgt.segFrames)+1)
		superimp = SuperimposeTracker(tgt.lengthInFrames, len(tgt.segFrames), ops.SUPERIMPOSE.overlapAmpThresh, len(ops.CORPUS), p)
		segs = []
		for start, end in tgt.segFrames: segs.append( TgtSegment(start, end) )
		######################################
		## sort segments by power if needed ##
		######################################
		if ops.SUPERIMPOSE.searchOrder == 'power':
			segs = sorted(segs, key=operator.attrgetter('power'), reverse=True)
		#########################
		## TARGET SEGMENT LOOP ##
		#########################
		for segidx, seg in enumerate(segs):
			tgt.timeInFrames = seg.start
			while True:
				##############################################################
				## check to see if we are done with this particular segment ##
				##############################################################
				if tgt.timeInFrames >= seg.end: break 
				########################################
				## run selection superimposition test ##
				########################################
				segidxt = superimp.test('segidx', segidx, ops.SUPERIMPOSE.minSegment, ops.SUPERIMPOSE.maxSegment)
				overt = superimp.test('overlap', tgt.timeInFrames, ops.SUPERIMPOSE.minOverlap, ops.SUPERIMPOSE.maxOverlap)
				onsett = superimp.test('onset', tgt.timeInFrames, ops.SUPERIMPOSE.minOnset, ops.SUPERIMPOSE.maxOnset)
				trig = tgt.thresholdTest(tgt.timeInFrames)
				trigVal = utilities.ampToDb(tgt.data['raw']['power'][tgt.timeInFrames])
				####################################################
				# skip selecting if some criteria doesn't match!!! #
				####################################################
				if 'notok' in [onsett, overt, segidxt]:
					if segidxt == 'notok':
						superimp.skip('maximum selections this segment', superimp.cnt['segidx'][segidx])
					if onsett == 'notok':
						superimp.skip('maximum onsets at this time', superimp.cnt['onset'][tgt.timeInFrames])
					if overt == 'notok':
						superimp.skip('maximum overlapping selections', superimp.cnt['overlap'][tgt.timeInFrames])
					tgt.timeInFrames += ops.SUPERIMPOSE.incr
					continue # next frame
				##############################################################
				## see if a selection should be forced without thresholding ##
				##############################################################
				if 'force' not in [onsett, overt, segidxt]: # test for amplitude threshold
					if not trig:
						superimp.skip('target too soft', trigVal)
						tgt.timeInFrames += ops.SUPERIMPOSE.incr
						continue # not loud enough, next frame
				####################################
				## get target segment information ##
				####################################
				thisSegmentLength = tgt.segdata.getraw('effDur-seg', tgt.timeInFrames, None)	
				power = tgt.segdata.getraw('power-seg', tgt.timeInFrames, tgt.timeInFrames+thisSegmentLength)	
				peak = metrics.peakTimeSeg(tgt.data[ops.CONCATENATE_PEAK_ALIGN_ENVELOPE]['power'][tgt.timeInFrames:tgt.timeInFrames+thisSegmentLength])[0]
				##############################
				## get valid corpus handles ##
				##############################
				validHandles, validSteams = src.evaluateValidSamples(ops.MAX_ACTIVE_VOICES, superimp)
				if len(validHandles) == 0:
					superimp.skip('no corpus sounds made it past restrictions and limitations', None)
					tgt.timeInFrames += ops.SUPERIMPOSE.incr
					continue
				################################################
				## search and see if we find a winning sample ##
				################################################
				searchObj.setCorpus(validHandles, validSteams)
				returnBool = searchObj.executeSearch(ops.SEARCH, tgt, tgt.timeInFrames, thisSegmentLength, peak, tgt.segmentsPerFrame[tgt.timeInFrames], segidx+1, p)
				if not returnBool: # nothing valid, so skip to new frame...
					superimp.skip('no corpus sounds made it through the search passes', None)
					tgt.timeInFrames += ops.SUPERIMPOSE.incr
					continue
				###################################################
				## if passing this point, picking a corpus sound ##
				###################################################
				superimp.pick(trig, trigVal, onsett, overt, segidxt)
				outputFiles.minMaxSearch.append(searchObj.searchMinMax[-1])
				chosenSample, chosenData = searchObj.searchResults[-1]
				##### MODIFY CHOSEN SAMPLES AMPLITUDE? #####
				if ops.POST_SELECTION_AMP_SCALE in ["lstSqr", "lstsqr"]:
					sourceAmpScale, outputDb = tgt.leastSqrsAmp(chosenSample, ops.POST_SELECTION_AMP_MIN, ops.POST_SELECTION_AMP_MAX, chosenData['segmentSt'], chosenData[ops.SUPERIMPOSE.ampDur])
				elif ops.POST_SELECTION_AMP_SCALE in ["peakAmp"]:
					sourceAmpScale, outputDb = tgt.peakAmpCorrect(chosenSample, ops.POST_SELECTION_AMP_MIN, ops.POST_SELECTION_AMP_MAX, chosenData['segmentSt'], chosenData[ops.SUPERIMPOSE.ampDur])
				elif ops.POST_SELECTION_AMP_SCALE in ["averageAmp"]:
					sourceAmpScale, outputDb = tgt.averagedAmpCorrect(chosenSample, ops.POST_SELECTION_AMP_MIN, ops.POST_SELECTION_AMP_MAX, chosenData['segmentSt'], chosenData[ops.SUPERIMPOSE.ampDur])
				else: sourceAmpScale, outputDb = (1, 100)
				##### append the one that just got chosen..
				ag.selectedSegments.append((chosenData['simCalcStEnd'][0], chosenData['simCalcStEnd'][0]+chosenSample.lengthInFrames, chosenSample, sourceAmpScale))
				##### subtract power to update onset detection
				tgt.subtractPowerForOnsetEnvelope(chosenSample, sourceAmpScale, ops.SUPERIMPOSE.subtractScale, chosenData['segmentSt'], int(chosenData[ops.SUPERIMPOSE.subtractDur]))
				##### mix chosen sample's descriptors into a tgt dict that holds cumulative mixtures..
				if ops.SUPERIMPOSE.calcMethod == "mixture":
					tgt.mixSelectedSamplesDescriptors(chosenSample, sourceAmpScale, chosenData['segmentSt'], chosenData[ops.SUPERIMPOSE.subtractDur])
				#####
				tgt.selectedSegs.append((ag.f2s(tgt.timeInFrames), ag.f2s(thisSegmentLength), tgt.lastSegmentInitiated, tgt.lastSegmentTerminated))
				trans = utilities.getTransposition(tgt, chosenSample, thisSegmentLength, chosenSample.transQuantize)
				if chosenData['trans'] != 0.: trans = chosenData['trans']
				##### manage loudestPeak descriptor
				if tgt.data['raw'].has_key('loudestPeak'):
					currentPeak = int(np.median(tgt.segmentsPerFrame[tgt.timeInFrames:tgt.timeInFrames+thisSegmentLength]))+1
					tgt.data['raw']['loudestPeak'][tgt.timeInFrames:tgt.timeInFrames+thisSegmentLength] = tgt.data['raw']['peakFrq-'+str(currentPeak%ops.NUMBER_OF_SPECTRAL_PEAKS)][tgt.timeInFrames:tgt.timeInFrames+thisSegmentLength]
	
				if ops.CSOUND_SCORE_DURATION_TYPE == 'corpusDur':
					durSec = chosenSample.endSec-chosenSample.startSec
				else:
					durSec = ag.f2s(chosenData[ops.CSOUND_SCORE_DURATION_TYPE])
				outputFiles.addEvent(chosenSample, utilities.limit(ag.f2s(chosenData['segmentSt']), 0, None), durSec, ag.f2s(thisSegmentLength), outputDb, trans, segidx, superimp.cnt['overlap'][tgt.timeInFrames], ops.CSOUND_STRETCH_CORPUS_DURATIONS_TO_MATCH_TARGET)
	
				minSelectedDuration = min(thisSegmentLength, chosenData['corpusEffDur'])
				tgt.segmentsPerFrame[tgt.timeInFrames:tgt.timeInFrames+minSelectedDuration] += 1
				if ops.CORPUS_POCKETING:
					src.updateTargetPocketCorpusRestriction(tgt.timeInFrames, tgt.segFrames, chosenSample)
				
				tgt.onsetsPerFrame[tgt.timeInFrames] += 1
				src.corpusVcOverlapCounter[chosenSample.voiceID][tgt.timeInFrames:tgt.timeInFrames+chosenData[ops.SUPERIMPOSE.overlapDur]] += 1
				src.lastVoice = chosenSample.voiceID
				src.updateSourcePool(chosenSample)
				src.updateRestrict(chosenSample, chosenData)
				superimp.increment(tgt.timeInFrames, thisSegmentLength, segidx, chosenSample.voiceID, chosenSample.data['raw']['power'])
		return error_accum





		

	def render(self, outputFileName=None):
		#print ag.selectedSegments
		global ag, tgt, outputFiles, ops, settingsFile, src, p
		if ops.CSOUND_RENDER_FILEPATH == None: return False
		p.logsection('RENDER')
		ag.timer("render", "set")
		if utilities.which('csound') == None:
			print "\n\tCannot render file because cannot find csound.\n"
			return False

		if outputFileName != None: ag.scoresToRender[0].outputName = outputFileName # user force
		renderWithCsound(ag.scoresToRender, "Rendering "+str(len(ag.scoresToRender))+" file(s) with Csound")
		# other stuff
		if ops.CSOUND_OUTPUT_MIX_FILEPATH != None:
			p.post("Making Csound Mix...", 1)
			utilities.executeCommand(['csound --output="'+ops.CSOUND_OUTPUT_MIX_FILEPATH+'"  "'+os.path.join(ops.AG_PATH,'scripts','csoundMix.orc')+'" "'+outputFiles.scoreMixHandleName+'" > /tmp/out  2>&1'])
		dateString='%s.%s-%s:%s:%s'%(localtime()[1], localtime()[2], localtime()[3], localtime()[4], localtime()[5])
		utilities.executeCommand(['echo "'+ag.scoresToRender[0].outputName+'" > /tmp/last_paplay']) # for my play script...
		ag.timer("render", "get")
		if ops.PRINT_ELAPSED_TIMES: ag.printTimeData()
		ag.scoresToRender = []
		p.done()
		return ops.CSOUND_PLAY_RENDERED_FILE

