import numpy as np
np.seterr(invalid='ignore')
import sys, os
import utilities
import simcalc


try:
	from aubio.aubioclass import *
except ImportError, err:
	pass

def set_all_segmented_methods(list):
	global ALL_SEG_METHDS
	ALL_SEG_METHDS = list



##############################
## SINGLE-NUMBER MORPHOLOGY ##
##############################
def peakTimeSeg(powers):
	return [np.argmax(powers)+0.5] # the middle of the peak window, reduces error to +/-0.5




def logAttackTime(powers, ag):
	idx = np.argmax(powers)
	time = np.clip(ag.f2s(idx), 0.0001, sys.maxint)
	return [np.log(time)] # the middle of the peak window, reduces error to +/-0.5




def f0Seg(f0s):
	idxs = np.nonzero(f0s)[0]
	tmp = []
	for idx in idxs: tmp.append(f0s[idx])
	if len(tmp) == 0: return [0.] # 0 means f0 didn't find an estimate!
	return [np.median(tmp)]




def descriptorSlope(handle, descriptName, start, NORMALIZE_FOR_DURATION=True):
	# with Norbert Schnell
	if start == None: effStart = 0
	else: effStart = start
	effDur = handle.segdata.getraw('effDur-seg', effStart, None)
	descript = handle.data['raw'][descriptName][effStart:effStart+effDur]
	start = int((len(descript))*0.2) # only take descriptor from %20-%80 
	end = int((len(descript))*0.8)
	values = np.array(descript[start:end])
	if len(values) < 2: return [0.0] # nada
	if NORMALIZE_FOR_DURATION:
		cmpLine = np.linspace(0, 1, num=len(values))
	else:
		cmpLine = np.linspace(0, len(values), num=len(values))
	return [np.polyfit(cmpLine, values, 1)[0]]




def MIDIPitchByFileName(name, midiPitchMethod, handle, notfound=-1):
	# midiPitchMethod == 'filename' || int/float || yin-seg || centroid-seg
	if midiPitchMethod == 'filename':
		fileHead = os.path.splitext(os.path.split(name)[1])[0]
		test = getMidiPitchFromString(fileHead)
		if test != False: return test
		else: return notfound
	elif midiPitchMethod == 'composite':
		test = getMidiPitchFromString( os.path.splitext(os.path.split(name)[1])[0] )
		if test != False:
			return test
		else:
			return MIDIPitchByFileName(name, 'f0-seg', handle)
	elif type(midiPitchMethod) in [float, int]:
		return midiPitchMethod
	elif midiPitchMethod == 'centroid-seg':
		return utilities.frq2Midi(handle.segdata.getraw('centroid-seg', 0, None))
	elif midiPitchMethod == 'f0-seg':
		if handle.segdata.getraw('f0-seg', 0, None) != 0:
			return utilities.frq2Midi(handle.segdata.getraw('f0-seg', 0, None))
		else:
			return notfound



FILENAME_PITCH_DICT = {} # save string results here so it doesn't have to reevaulate..
def getMidiPitchFromString(string):
	import os, re
	global FILENAME_PITCH_DICT
	if FILENAME_PITCH_DICT.has_key(string): return FILENAME_PITCH_DICT[string]
	pitchDict = {"c": 0,"cs": 1,"db": 1,"d": 2,"ds": 3,"eb": 3,"e": 4, 'fb': 4, "f": 5,"fs": 6,"gb": 6,"g": 7,"gs": 8,"ab": 8,"a": 9,"as": 10,"bb": 10, "b": 11, 'cb': 11}
	results = []
	results.extend(re.compile(r'[a-gA-G][sb\#][0123456789]').findall(string)) # re command for searching three element pattern
	results.extend(re.compile(r'[a-gA-G][0123456789]').findall(string)) # no accidentals!!
	if len(results) > 0: # found something
		result = results[0]
		if len(result) == 2: 
			pch = result[0].lower()
			octv = result[1]
		else:
			pch = result[0:2].lower().replace('#', 's')
			octv = result[2]
		returnPitch = pitchDict[pch]+(int(octv)+1)*12
		FILENAME_PITCH_DICT[string] = returnPitch
		return returnPitch
	else:
		FILENAME_PITCH_DICT[string] = False
		return False # could'nt find one
	


def percentInFile(handle, start, end):
	if handle.typed == 'target':
		output = [(start/float(handle.lengthInFrames))*100] # computed in frames
	else:
		output = [(handle.startSec/float(readsf.fileDuration(handle.filename)))*100] # computed in seconds
	return output



def effectiveDur(handle, time, ag):
	# RETURN SEGMENT LENGTH IN FRAMES
	if time == None: time = 0
	length = len(handle.data['raw']['power'])
	if handle.typed == 'target':
		start = time+ag.minSegLength
		if start >= length:  # if shorter than user specified segment time...
			handle.lastSegmentTerminated = 'fileEnd'
			return length-time
		else:
			stop = start+1
			if ag.minSegLength > 1: # get max in this time
				maxPowerInMinTime = max(handle.data[ops.TARGET_OFFSET_ENVELOPE]['power'][time:time+ag.minSegLength])
			else: # just use first value
				maxPowerInMinTime = handle.data[ops.TARGET_OFFSET_ENVELOPE]['power'][time]
			# drop!!!! important
			while stop+1 < length:
				div = np.clip(handle.data[ops.TARGET_OFFSET_ENVELOPE]['power'][stop], 0.00000001, sys.maxint)
				riseratio = handle.data[ops.TARGET_OFFSET_ENVELOPE]['power'][stop+1]/div
				if handle.data[ops.TARGET_OFFSET_ENVELOPE]['power'][stop] <= ag.powerOffsetValue:
					stop += 1 # use next frame too, since looking ahead...
					handle.lastSegmentTerminated = ['drop', utilities.trunc(handle.data[ops.TARGET_OFFSET_ENVELOPE]['power'][stop], 2), str(ag.powerOffsetValue)]
					break
				elif riseratio >= handle.rise_ratio_offset and handle.thresholdTest(stop+1):
					handle.lastSegmentTerminated = ['rise', utilities.trunc(handle.data[ops.TARGET_OFFSET_ENVELOPE]['power'][stop+1]/handle.data[ops.TARGET_OFFSET_ENVELOPE]['power'][stop], 2), str(handle.rise_ratio_offset)]
					break
				else:
					stop += 1 # keep looking
			if stop-time > ag.maxSegLength: # if longer than user specified maximum segment time...
				handle.lastSegmentTerminated = 'maxSegLength'
				return ag.maxSegLength
			else:
				# return the stopping point
				return stop-time
	################
	## for corpus ##
	################
	elif handle.typed == 'source':
		counter = length-1
		while True: # count from the end
			if handle.data['raw']['power'][counter] > ag.absThreshScalar: break
			counter -= 1
			if counter < 0: return handle.lengthInFrames
		realDur = handle.endSec-handle.startSec
		time = ops.DESCRIPTOR_HOP_SIZE_SEC*(counter+1)
		
#		if realDur-time > 1:
#			print counter+1, handle.endSec-handle.startSec
#			print handle.name
#			print handle.data['raw']['power']
#			print counter+1, handle.endSec-handle.startSec
		return counter+1



def delta(vals):
	if len(vals) > 1:
		output = np.diff(vals)
		output = np.insert(output, -1, 0)
		return output
	else:
		return vals


def odf(data, numberAvg): 
	# with Norbert Schnell
	# 	an averaged delta function
	dataLen = len(data)
	matrix = np.zeros((numberAvg,dataLen)) # matrix of length x number average
	for frame in range(len(data)):
		for i in range(1,numberAvg+1):
			if frame-i >= 0: matrix[i-1][frame] = data[frame-i] # fill it
	medians = np.median(matrix, axis=0)
	return data-medians




def hannWin(size):
	n = np.arange(size)
	w = 0.5*(1.0-np.cos(2*np.pi*n/(float(size-1))))
	return w



########################
## DESCRIPTOR MIXTURE ##
########################
def mixTwoTimeVaryingFeatures(descript, tgtTime, mixDur, mixPointer, srcPointer):
	if descript.find('power') != -1 or descript.find('rms') != -1: # powers are summed, not averaged
		# (e1+e2)
		#print "adding powers"
		#print "\t", mixPointer[descript][tgtTime:tgtTime+mixDur]
		#print "\t", srcPointer[descript][0:mixDur]
		#print "\t=", mixPointer[descript][tgtTime:tgtTime+mixDur]+srcPointer[descript][0:mixDur]
		return mixPointer[descript][tgtTime:tgtTime+mixDur]+srcPointer[descript][0:mixDur]
	else:
		# ((d1*e1)+(d1*e1)) / (e1+e2)
#		print "mixing descriptors"
#		print "time", tgtTime, "mix duration", mixDur, len(mixPointer[descript]), len(srcPointer[descript])
#		print "\t", mixPointer[descript][tgtTime:tgtTime+mixDur]
#		print "\t", srcPointer[descript][0:mixDur]
#		print "\t=", descript, ((mixPointer[descript][tgtTime:tgtTime+mixDur]*mixPointer['power'][tgtTime:tgtTime+mixDur])+(srcPointer[descript][0:mixDur]*srcPointer['power'][0:mixDur]))/(mixPointer['power'][tgtTime:tgtTime+mixDur]+srcPointer['power'][0:mixDur])
		return ((mixPointer[descript][tgtTime:tgtTime+mixDur]*mixPointer['power'][tgtTime:tgtTime+mixDur])+(srcPointer[descript][0:mixDur]*srcPointer['power'][0:mixDur]))/(mixPointer['power'][tgtTime:tgtTime+mixDur]+srcPointer['power'][0:mixDur])


def mixTwoSegmentedFeatures(descript, feature1, power1, feature2, power2):
	#print descript, feature1, power1, feature2, power2
	if descript in ['power-seg', 'rms-seg']:
		return power1+power2
	else: # weighted average
		return ((feature1*power1)+(feature2*power2))/(power1+power2)




#############################
## PACKAGES OF DESCRIPTORS ##
#############################
def name2packageExpansion(name, weight, neededBy):
	metricsToWrite = []
	packagename = name

	#1-11, 12-23
	if name.find('mfccs') != -1:
		mfccWeightMask = np.power(np.linspace(1., 0.05, num=ops.IRCAMDESCRIPTOR_NUMB_MFCCS-1), 0.5)
		for i in range(ops.IRCAMDESCRIPTOR_NUMB_MFCCS-1): metricsToWrite.append(('mfcc'+str(i+1)+name[5:], weight*(mfccWeightMask[i]/sum(mfccWeightMask)), neededBy ))
	elif name.find('chromas') != -1:
		for i in range(12): metricsToWrite.append( ( 'chroma'+str(i)+name[9:], weight/12., neededBy ) )
	elif name.find('autocorrs') != -1:
		for i in range(12): metricsToWrite.append( ( 'autocorr'+str(i)+name[9:], weight/12., neededBy ) )
	elif name.find('cluster_cents') != -1:
		numb = len(ops.CLUSTERANAL_DESCRIPTOR_DIM)
		for i in range(numb):
			metricsToWrite.append( ( 'cluster_cent'+str(i)+name[13:], weight/float(numb), neededBy ) )
		#print metricsToWrite
	elif name.find('peakfrqs') != -1:
		for i in range(ops.SUPERVP_NUMB_PEAKS):
			metricsToWrite.append( ('peakfrq'+str(i)+name[10:], weight*((ops.SUPERVP_NUMB_PEAKS-i)/float(ops.SUPERVP_NUMB_PEAKS)), neededBy ) )
			metricsToWrite.append( ('peakamp'+str(i)+name[10:], 0, neededBy) )
#	elif name.find('loudestpeak') != -1:
#		metricsToWrite.append( ('peakfrq0'+name[10:], weight*((ops.SUPERVP_NUMB_PEAKS-i)/float(ops.SUPERVP_NUMB_PEAKS)), neededBy ) )
#		metricsToWrite.append( ('peakamp0'+name[10:], 0, neededBy) )
#		for i in range(ops.SUPERVP_NUMB_PEAKS+1):
#			metricsToWrite.append( ('peakfrq'+str(i)+name[10:], weight*((ops.SUPERVP_NUMB_PEAKS-i)/float(ops.SUPERVP_NUMB_PEAKS)), ['target'] ) )
#			metricsToWrite.append( ('peakamp'+str(i)+name[10:], 0, ['target']) )
	elif name.find('crests') != -1:
		for i in range(4):
			metricsToWrite.append( [ 'crest'+str(i)+name[6:], weight/4., neededBy ] )
	elif name.find('flatnesses') != -1:
		for i in range(4):
			metricsToWrite.append( [ 'flatness'+str(i)+name[10:], weight/4., neededBy ] )
	elif name.find('harmonictristimuluses') != -1:
		for i in range(3):
			metricsToWrite.append( [ 'harmonictristimulus'+str(i)+name[21:], weight/3., neededBy ] )
	elif name.find('perceptualtristimuluses') != -1:
		for i in range(3):
			metricsToWrite.append( [ 'perceptualtristimulus'+str(i)+name[23:], weight/3., neededBy ] )
	else:
		metricsToWrite = [(name, weight, neededBy)]
		packagename = None
	return packagename, metricsToWrite




##############################
## AUDIOGUIDE FUNCTION CALL ##
##############################
def analSegment(d, handle, start, end, typed, ag):
	global ALL_SEG_METHDS
	# if already computed as init, skip it! -- really a hack in case there is a bug -- shouldn't happen
	if typed != "Recalc" and start == None and end == None and handle.data['raw'].has_key(d.name): pass
	#print d.name, d.seg
	if d.seg:
		# SINGLE-NUMBER METHODS
		if d.name == "dur-seg": # raw segment duration
			output = [handle.lengthInFrames]
		elif d.name == "effDur-seg": # perceived segment duration
			output = [effectiveDur(handle, start, ag)]
		elif d.name == "MIDIPitch-seg":
			output = [MIDIPitchByFileName( handle.filehead, handle.midiPitchMethod, handle)]
		elif d.name == "peakTime-seg":
			output = peakTimeSeg(handle.data['raw']['power'][start:end])
		elif d.name == "logAttackTime-seg":
			output = logAttackTime(handle.data['raw']['power'][start:end], ag)
		elif d.name == "percentInFile-seg":
			output = percentInFile(handle, start, end)
		elif d.name == "f0-seg":
			output = f0Seg(handle.data['raw']['f0'][start:end])
		# DESCRIPTOR SLOPE
		elif d.type == 'slope-regression':
			output = descriptorSlope(handle, d.parents[0], start)
		# FLAT MEAN - NOT ENERGY WEIGHTED
		elif d.type  == 'mean':
			output = [np.average(handle.data['raw'][d.parents[0]][start:end])]
		# SEGMENT-AVERAGE METHODS
		elif d.seg_method  == 'max':
			output = [np.max(handle.data['raw'][d.parents[0]][start:end])]
		elif d.seg_method  == 'median':
			output = [np.median(handle.data['raw'][d.parents[0]][start:end])]
		else: # 'weighted_mean'
			try:
				output = [np.average(handle.data['raw'][d.parents[0]][start:end], weights=handle.data['raw']['power'][start:end])]
			except ZeroDivisionError: # in case all powers are zero
				output = [np.average(handle.data['raw'][d.parents[0]][start:end])]

	else:
		# DELTA
		if d.type == 'delta':
			output = delta(handle.data['raw'][d.parents[0]][start:end])
		# DELTA DELTA
		elif d.type == 'deltadelta':
			output = delta(handle.data['raw'][d.parents[0]][start:end])
			output = delta(output)
		# ONSET DETECTION FUNCTION CALL
		elif d.type == 'onsetdetection':
			analSplit = d.name.split('-')
			output = odf(handle.data['raw'][d.parents[0]][start:end], int(analSplit[2]))
		elif d.name == "loudestpeak":
			output = handle.data['raw']['peakfrq0'][start:end]
		else:
			print "No method for", anal, ", Quitting..."
			utilities.exit()
	return output


def multiSegAnal(handle, descript, start, end):
	if start == None: start = 0
	p = int(handle.data['raw']['peakTime-seg'][start])
	output = np.zeros(3)
	output[0] = [simcalc.energyWeightSingle((handle.data['raw'][descript][start:p-1]), handle.data['raw']['power'][start:p-1])][0]
	output[1] = [simcalc.energyWeightSingle((handle.data['raw'][descript][p:p+1]), handle.data['raw']['power'][p:p+1])][0]
	output[2] = [simcalc.energyWeightSingle((handle.data['raw'][descript][p+2:end]), handle.data['raw']['power'][p+2:end])][0]
	return output
